
#include "graphics.h"
#include "spritestate.h"
#include "game.h"
#include "JAGUAR.H"

/* --------------------------------------------------------------------------*/
/* Constants */

/* Bitmaps positions in game object list */
#define LAVA_OFFSET   0
#define LAVA_N        6

#define JUNK_OFFSET   LAVA_OFFSET+LAVA_N
#define JUNK_N        5

#define MISSILE_OFFSET JUNK_OFFSET+JUNK_N
#define MISSILE_N     6

#define PLAYER_OFFSET MISSILE_OFFSET+MISSILE_N
#define PLAYER_N      1

#define ENEMY_OFFSET  PLAYER_OFFSET+PLAYER_N
#define ENEMY_N       7

#define SCOREWINDOW_OFFSET ENEMY_OFFSET+ENEMY_N
#define SCOREWINDOW_N 1

#define MESSAGEWINDOW_OFFSET SCOREWINDOW_OFFSET+SCOREWINDOW_N
#define MESSAGEWINDOW_N 1

#define LOGWINDOW_OFFSET  MESSAGEWINDOW_OFFSET+MESSAGEWINDOW_N
#define LOGWINDOW_N    1

#define N_GAME_BITMAP_OBJ  LAVA_N+JUNK_N+MISSILE_N+PLAYER_N+ENEMY_N+SCOREWINDOW_N+MESSAGEWINDOW_N+LOGWINDOW_N


/* Bitmaps position in menu object list */
#define TITLE_OFFSET  LAVA_OFFSET+LAVA_N
#define TITLE_N       1

#define CREDITS_OFFSET  TITLE_OFFSET+TITLE_N
#define CREDITS_N     1

#define U235BM_OFFSET  CREDITS_OFFSET+CREDITS_N
#define U235BM_N      1

#define N_MENU_BITMAP_OBJ LAVA_N+TITLE_N+CREDITS_N+U235BM_N


/* Sizes */
#define LAVA_W 80
#define LOGH 64
#define MESSAGEH 16
#define SCOREH 16
#define CREDITSH 64

/* --------------------------------------------------------------------------*/
/* Externs */
extern unsigned short font8x8;
extern unsigned short lavabitmapdata;
extern unsigned short messagewindowbase;
extern unsigned short scorewindowbase;
extern unsigned short titlebitmapdata;
extern unsigned short creditwindowbase;
extern unsigned short u235sebitmapdata;

/* --------------------------------------------------------------------------*/
/* Forward declaration */

void ScrollLava(CBitmap* );

/* --------------------------------------------------------------------------*/
/* Game globals */
TGameStats gamestats;


/* --------------------------------------------------------------------------*/
/* Menu loop */

/* Menu bitmaps list */
CBitmap menuBitmaps[N_MENU_BITMAP_OBJ];
char menulistbuff[DLSIZE(N_GAME_BITMAP_OBJ)+16];  /* Object list buffer */
static int posx = 0;

/* Menu loop
 * Display credits and greetings.
 * Wait for user to select a level.
 * 
 * OUT: (int) level selected   
 */
int MenuLoop()
{
  CBitmapArray menuarray;
  void* ptlist;
  int i;
  
  int selection = 0;
  
  ptlist = ALIGN(menulistbuff);
  
  Memset64(&creditwindowbase, 0, 0, 256*CREDITSH*2);
   
  /* Init lava bitmaps */
	for(i = 0; i < LAVA_N; i++)
	{
    MakeBitmap(&menuBitmaps[LAVA_OFFSET+i], &lavabitmapdata, LAVA_W*i+(posx>>1), 0, LAVA_W, FBH, 0);
	}	
	/* Init title bitmap */
	MakeBitmap(&menuBitmaps[TITLE_OFFSET], &titlebitmapdata, (FBW-256)/2, 0, 256, 64, O_TRANS);
	
	/* Init credit window */
	MakeBitmap(&menuBitmaps[CREDITS_OFFSET], &creditwindowbase, (FBW-256)/2, 76, 256, CREDITSH, O_TRANS);
	
	/* Init U235se bitmap */
	MakeBitmap(&menuBitmaps[U235BM_OFFSET], &u235sebitmapdata, (FBW-192)/2, FBH-76, 192, 75, O_TRANS);
	
	InitBitmapArray(&menuarray, ptlist, &menuBitmaps[0], N_MENU_BITMAP_OBJ);	
	ActiveBitmapArray = &menuarray;
	
	LibDrawInit(&creditwindowbase, WID256, CREDITSH);
  SelectFont(&font8x8, 8,8);
     
  DrawText(28,0,"Code & Design by DrTypo.");
  DrawText(16,10,"Press 1..9 to start a level.");
  DrawText(8,30,"Greetings to Atari Age people!");
  
	while(!selection)
	{
    WaitVBL();	
    ScrollLava(&menuBitmaps[LAVA_OFFSET]);
		ReadJoypads();
		if( (joy1cur & (1<<KEY_1))
		    &(joy1edge & (1<<KEY_1)) )
		    selection = 1;
		else if( (joy1cur & (1<<KEY_2))
		    &(joy1edge & (1<<KEY_2)) )
		    selection = 2;
		else if( (joy1cur & (1<<KEY_3))
		    &(joy1edge & (1<<KEY_3)) )
		    selection = 3;
		else if( (joy1cur & (1<<KEY_4))
		    &(joy1edge & (1<<KEY_4)) )
		    selection = 4;
		else if( (joy1cur & (1<<KEY_5))
		    &(joy1edge & (1<<KEY_5)) )
		    selection = 5;
		else if( (joy1cur & (1<<KEY_6))
		    &(joy1edge & (1<<KEY_6)) )
		    selection = 6;
    else if( (joy1cur & (1<<KEY_7))
		    &(joy1edge & (1<<KEY_7)) )
		    selection = 7;
    else if( (joy1cur & (1<<KEY_8))
		    &(joy1edge & (1<<KEY_8)) )
		    selection = 8;   
    else if( (joy1cur & (1<<KEY_9))
		    &(joy1edge & (1<<KEY_9)) )
		    selection = 9;	  
	}
	return selection;
}


/* --------------------------------------------------------------------------*/
/* Game loop */


/* Game bitmaps list */
CBitmap gamebitmaps[N_GAME_BITMAP_OBJ]; /* Bitmaps structures used to create and refresh the object list */
char gamelistbuff[DLSIZE(N_GAME_BITMAP_OBJ)+16];  /* Object list buffer */

/* Sprites */
CSpriteState player;
CSpriteState enemy[ENEMY_N];
CSpriteState missile[MISSILE_N];

/* Game loop
 * Handle game logic: move player and enenmies, collision detection...
 *
 * IN:  startlevel (int)  Difficulty level at which the game starts.
 */
void GameLoop(int startlevel)
{
  CBitmapArray gamearray;  
  void* ptgamelist;
	int i;
  short d;	
  int paused = 0;
	unsigned long pauseticks = 0;
	int endgame = 0;
	int displaylog = 0;
			
	gamestats.nbkilled = 0;
	gamestats.score = 0;
	gamestats.level = startlevel;
	gamestats.lives = 3;
	
	srand(ticks);
	
	/* Clear windows */
	Memset64(&logwindowbase, 0, 0, 256*LOGH*2);
	Memset64(&messagewindowbase,0 ,0, 80*MESSAGEH*2);
	Memset64(&scorewindowbase, 0, 0, 320*SCOREH*2);
	
	/* 16 byte aligned pointer to object list */
	ptgamelist = ALIGN(gamelistbuff);
	
	/* Init lava bitmaps */
	for(i = 0; i < LAVA_N; i++)
	{
    MakeBitmap(&gamebitmaps[LAVA_OFFSET+i], &lavabitmapdata, LAVA_W*i+(posx>>1), 0, LAVA_W, FBH, 0);
	}	
	/* Init junk bitmap */
	for(i = 0; i < JUNK_N; i++)
	{
	   MakeBitmap(&gamebitmaps[JUNK_OFFSET+i], &junkbitmapdata, -JUNK_W,412, JUNK_W, JUNK_H, 0);
	   gamebitmaps[JUNK_OFFSET+i].csflags |= (1<<ACTIVE_COLLISION_FLAG);
  }  
	/* Init missiles bitmap */
	for(i = 0; i < MISSILE_N; i++)
	{
    MakeBitmap(&gamebitmaps[MISSILE_OFFSET+i], &missilebitmapdata, 10, 50, MISSILE_W, MISSILE_H, O_TRANS);
	}  	
	/* Init player bitmap */
	MakeBitmap(&gamebitmaps[PLAYER_OFFSET], &playerbitmapdata, 0, 0, PLAYER_W, PLAYER_H, O_TRANS);	
	/* Init enemies bitmap */
	for(i = 0; i < ENEMY_N; i++)
	{
	 MakeBitmap(&gamebitmaps[ENEMY_OFFSET+i],  &enemybitmapdata, FBH, 100, PLAYER_W, PLAYER_H, O_TRANS);
	}
  /* Init score window bitmap */
  MakeBitmap(&gamebitmaps[SCOREWINDOW_OFFSET], &scorewindowbase, 0, 0, 320, SCOREH, O_TRANS);
  LibDrawInit(&scorewindowbase, WID320, 16);
  SelectFont(&font8x8, 8,8);
  DrawText(0,0,"Lives:");
  DrawText(104,0, "Score:");
  DrawText(248, 0, "Level:");
  
  /* Init message window bitmap */
  MakeBitmap(&gamebitmaps[MESSAGEWINDOW_OFFSET], &messagewindowbase, (FBW-80)/2, 400, 80, MESSAGEH, O_TRANS);
    
	/* Init log window bitmap */
	MakeBitmap(&gamebitmaps[LOGWINDOW_OFFSET], &logwindowbase, 0, 400, 256, LOGH, O_TRANS);	
	LibDrawInit(&logwindowbase, WID256, LOGH);
  SelectFont(&font8x8, 8,8);		
	DrawText(0,0,"Log");
  DrawHex(0,8,*HDB1);
  DrawHex(9*8,8,*HDB2);
  DrawHex(18*8,8,*HDE);
  
  /* Create and activate object list */
  InitBitmapArray(&gamearray, ptgamelist, &gamebitmaps[0], N_GAME_BITMAP_OBJ);	
	ActiveBitmapArray = &gamearray;
	
  /* Init sprites */  
  MakeSpriteState(&player, &gamebitmaps[PLAYER_OFFSET], EAlive, EPlayer);
  for(i = 0; i < ENEMY_N; i++)
  {
    MakeSpriteState(&enemy[i], &gamebitmaps[ENEMY_OFFSET+i], EDead, EEnemy);
  }
  for(i = 0; i < MISSILE_N; i++)
  {
    MakeSpriteState(&missile[i], &gamebitmaps[MISSILE_OFFSET+i], EDead, EMissile);
  }
  
  GeneratePlayer(&player);
   
  LibDrawInit(&scorewindowbase, WID320, SCOREH);
  SelectFont(&font8x8, 8,8);
      
  /* Game loop */ 
	while(!endgame)
	{
		WaitVBL();
		
		ReadJoypads();		
		
		/* Test paused */
		if( (joy1cur & (1<<PAUSE)) 
    &&  (joy1edge & (1<<PAUSE)) )
    {
      paused = !paused;
      if(paused)
      {
        gamebitmaps[MESSAGEWINDOW_OFFSET].Y = (FBH-MESSAGEH)/2;
        pauseticks = ticks;
      }
      else
      {
        gamebitmaps[MESSAGEWINDOW_OFFSET].Y = 400;
        ticks = pauseticks;
      }
     } 
		
		/* Test display log window */
	/*	if( (joy1cur & (1<<KEY_STAR)) 
    &&  (joy1edge & (1<<KEY_STAR)) )
    {
      displaylog = !displaylog;
    }      
		if(displaylog)
		{
		  gamebitmaps[LOGWINDOW_OFFSET].Y = 13;
    }
    else
    {
      gamebitmaps[LOGWINDOW_OFFSET].Y = 380;
    }*/
  
    /* Update score windows */  
    DrawDec(48,0,gamestats.lives,1);
    DrawDec(152,0, gamestats.score, 6);
    DrawDec(296, 0, gamestats.level, 2);
  
    /* Display GAME OVER message and test joypad to exit */
    if(gamestats.lives == 0)
    {
      paused = 0;
      gamebitmaps[MESSAGEWINDOW_OFFSET].Y = (FBH-MESSAGEH)/2;
    	LibDrawInit(&messagewindowbase, WID80, MESSAGEH);
      SelectFont(&font8x8, 8,8);	
      DrawText(4, 0, "GAME OVER"); 
      LibDrawInit(&scorewindowbase, WID320, SCOREH);
      SelectFont(&font8x8, 8,8);
      
     	if( (joy1cur & (1<<FIRE_B)) 
        &&  (joy1edge & (1<<FIRE_B)) )
      {
        endgame = -1;
      }
    }
    
		if(!paused)
		{
  		ScrollLava(&gamebitmaps[LAVA_OFFSET]);		
  		
  		GenerateEnemies(&enemy[0], ENEMY_N);
  		GenerateJunk(&gamebitmaps[JUNK_OFFSET], JUNK_N);
  		
  		MovePlayer(&player);
  		MoveEnemies(&enemy[0], ENEMY_N);
  		MoveMissiles(&missile[0], MISSILE_N);
  		MoveJunk(&gamebitmaps[JUNK_OFFSET], JUNK_N);
  		
  		TestPlayerFire(&player, &missile[0], MISSILE_N);
  		TestRespawnPlayer(&player);
  		
  		TestPlayerVsEnemyCollision(&player, &enemy[0], ENEMY_N);
  		TestPlayerVsJunkCollision(&player, &gamebitmaps[JUNK_OFFSET], JUNK_N);
  		TestMissileVsEnemyCollision(&missile[0], MISSILE_N, &enemy[0], ENEMY_N);
  		
    		
  		/* Give new life */
  		if((gamestats.lives > 0) && (gamestats.nbkilled > NEW_LIFE_REQ))
  		{
  		  gamestats.lives++;
  		  if(gamestats.lives > 9) gamestats.lives = 9;
  		  gamestats.nbkilled = 0;
      }
      
      /* Too many enemies escaped, game over */
      if(gamestats.score < 0)
      {
        gamestats.score = 0;
        gamestats.lives = 0;
        ExplodeSprite(&player);
      }
      else
      {
        int nlvl;              
        nlvl = DivU(gamestats.score,(100+gamestats.level*20));
                
        /* Compute new level */   		 
    		if((nlvl > gamestats.level) && (nlvl < 100))
    		{
     		  gamestats.level = nlvl;
        }
      }
      
  		UpdateSprites(&player, 1);
  		UpdateSprites(&enemy[0], ENEMY_N);
  	}
  	else
  	{  	   
  	  Memset64(&messagewindowbase,0 ,0, 80*MESSAGEH*2);
    	LibDrawInit(&messagewindowbase, WID80, MESSAGEH);
      SelectFont(&font8x8, 8,8);	
      DrawText(16, 0, "PAUSED"); 
      LibDrawInit(&scorewindowbase, WID320, SCOREH);
      SelectFont(&font8x8, 8,8);  	
    }
	}
}

/* Scroll lava background
 *
 * IN: bitmaps (CBitmap*) Pointer to lava bitmaps. 
 */ 
void ScrollLava(CBitmap* bitmaps)
{
  int i;
  if(bitmaps == NULL) return;
  
  for(i = 0; i < LAVA_N; i++)
  {
    bitmaps->X = (posx>>1) + i*LAVA_W;
    bitmaps++; 
  }
  posx--;
  if(posx <= -180) posx = -20;
}


