
#ifndef __GRAPHICS_H__
#define __GRAPHICS_H__

/*
 * Graphics Libs
 * U235 Sound Engine C wrapper   
 *  v0.0.27
 *
 */ 
   
#undef NULL
#define NULL ((void *)0)

/* Bitmaps structures */
typedef struct 
{
  void* addr;
  short X;
  short Y;
  short W;
  short H;
  unsigned short flags;
  unsigned short csflags;
  void* ptDL;
  unsigned long objhi;
  unsigned long objlo;  
} CBitmap; 

typedef struct
{
  unsigned short count;
  void* ptDL;
  CBitmap* bitmaps;
} CBitmapArray;

typedef struct
{
  CBitmap* bitmap;
  void* base;
  unsigned short increment;
  unsigned short steps;
  unsigned short ticks;
  unsigned short type;
  unsigned short stepCounter;
  unsigned short tickCounter;
} CAnim;

#define COLLIDE_FLAG  0
#define ACTIVE_COLLISION_FLAG 1

#define DLSIZE(X) (((X)<<4)+24)
#define ALIGN(X)   ((void*) ((((int)(X))+16) & 0xFFFFF0)) 
 
extern void WaitGPU();
extern void ConvertHex(int number, char* ptr);
extern void ConvertDec(int number, char* ptr, int digits);
extern void Memset64(void* dst, int pat1, int pat2, int len);
extern void Memcopy64(void* dst, void* src, int len);
extern void srand(int);
extern int  rand();
extern int  DivS(int dividende, int diviseur);
extern int  DivU(int dividende, int diviseur);
extern void LibDrawInit(void* videomemory, int blitwid, int vidh);
extern void SelectFont(void* font, int w, int h);
extern void DrawText(int x, int y, char* str);
extern void DrawHex(int x, int y, int n);
extern void DrawDec(int x, int y, int n, int digits);

extern void WaitVBL();

extern void MakeBitmap(CBitmap* dst, void* addr, int x, int y, int w, int h, int flags);
extern void InitBitmapArray(CBitmapArray* array, void* DL, CBitmap* bitmaps, int count);
extern int InitDL(void* ptDL, int dlsize);
extern int AddBitmap(char* ptDL, CBitmap* bitmapObj);
extern int AddStopObject(char* ptDL);
extern void TestBitmapCollision(CBitmap* probe, CBitmap* bitmaps, int n);

extern void MakeAnim(CAnim* dst, CBitmap* bitmap, int increment, int steps, int ticks);
extern void UpdateAnim(CAnim* dst, int n);
extern void ResetAnim(CAnim* dst, int n);

extern void ReadJoypads();

extern int  SetSample(int index, void* addr, void* addrend);
extern void PlaySample(int chan, int index);
extern void PlaySampleDoubleChan(int chan, int index);
extern void SetFreq(int chan, int freq);
extern void SetVol(int chan, int vol);
extern void ChangeSampleSign(int index);

extern unsigned long ticks;
extern unsigned long joy1edge;
extern unsigned long joy2edge;
extern unsigned long joy1cur;
extern unsigned long joy2cur;
extern CBitmapArray* ActiveBitmapArray;

#endif /* __GRAPHICS_H__ */
