
#include "jaguar.h"
#include "game.h"
#include "spritestate.h"


extern CSpriteState player;


/* Initialize a CSpriteState structure.
 * The CSpriteState will be linked to the CBitmap provided.
 * If the state is EDead, the sprite is hidden below the screen.
 * IN:   sprite (CSpriteState*) pointer to CSpriteState structure to initialize.
 *      bitmap (CBitmap*) pointer to bitmap to link.
 *      state (ESpriteState) initial state of the sprite.
 */ 
void MakeSpriteState(CSpriteState* sprite, CBitmap* bitmap, ESpriteState state, ESpriteType type)
{
  if(sprite == NULL) return;
  sprite->mBitmap = bitmap;
  sprite->mState = state;
  sprite->mType = type;
  sprite->mDirX = 0;
  sprite->mDirY = 0;
  if(bitmap != NULL)
  {
    if(state == EAlive)
      bitmap->csflags |= (1<<ACTIVE_COLLISION_FLAG);
    else
    {
      bitmap->csflags &= ~(1<<ACTIVE_COLLISION_FLAG); 
      bitmap->Y |= 512; /* Hide dead sprite */
    }
      
    bitmap->csflags &= ~(1<<COLLIDE_FLAG);    
  }
}

/* Move player if he's alive.
 * ReadJoypads() must be called in main loop for this procedure to work.
 * IN: player (CSpriteState*) pointer to player sprite.
 */ 
void MovePlayer(CSpriteState* player)
{
  int x,y;
  CBitmap* bitmap;
  
  if(player == NULL) return;
  if((player->mState != EAlive)&&(player->mState != EInvicible)) return;
  
  bitmap = player->mBitmap;
  if(bitmap == NULL) return;  
  
	if(joy1cur&(1<<JOY_UP))
	{		
    y = bitmap->Y;
    y -= PLAYER_SPEED_Y;
    if(y < 0) y = 0;
    bitmap->Y = y;   
	  
  }
  if(joy1cur&(1<<JOY_DOWN))
	{
	  y = bitmap->Y;
    y += PLAYER_SPEED_Y;
    if(y > (FBH-PLAYER_H)) y = (FBH-PLAYER_H);		
    bitmap->Y = y;   
  }
  if(joy1cur&(1<<JOY_LEFT))
  {
    x = bitmap->X;
    x -= PLAYER_SPEED_X;
    if(x < 0) x = 0; 
    bitmap->X = x;
  }
  if(joy1cur&(1<<JOY_RIGHT))
  {
    x = bitmap->X;
    x += PLAYER_SPEED_X;
    if(x > (FBW-PLAYER_W)) x = (FBW-PLAYER_W); 
    bitmap->X = x;
  }
}

/* Movie alive enemies.
 * IN: enemy (CSpriteState*) pointer to enemy sprite array.
 *      n (int) Number of sprites in array. 
 */
void MoveEnemies(CSpriteState* enemy,int n)
{
static int lastchange = 0;
  int i;
  CBitmap* bitmap;
  
  if(enemy == NULL) return;
  for(i = 0; i < n; i++)
  {
    if( (enemy->mState == EAlive) && ((bitmap=enemy->mBitmap) != NULL) )
    {
      int diry;
      int tmp;
      switch(enemy->mType)
      {
      case EEnemySpike:
        bitmap->X -=  enemy->mDirX;
        diry = enemy->mDirY;
        bitmap->Y += diry;
        if(bitmap->Y < 0)
        {
          bitmap->Y = 0;
          enemy->mDirY = -diry;
        }
        else if(bitmap->Y >= (FBH-PLAYER_H))
        {
          bitmap->Y = FBH-PLAYER_H-1;
          enemy->mDirY = -diry;
        }       
        break;
      case EEnemyCrab:   
        tmp = rand();     
        if((ticks - lastchange > 25) && ((tmp&0x3)==0))
        {
          lastchange = ticks;
          enemy->mDirY = -enemy->mDirY;
        }
        bitmap->X -=  enemy->mDirX;
        diry = enemy->mDirY;
        bitmap->Y += diry;
        if(bitmap->Y < 0)
        {
          bitmap->Y = 0;
          enemy->mDirY = -diry;
        }
        else if(bitmap->Y >= (FBH-PLAYER_H))
        {
          bitmap->Y = FBH-PLAYER_H-1;
          enemy->mDirY = -diry;
        }             
        break;
      case EEnemyCross:
        bitmap->X -= enemy->mDirX; 
        if(player.mBitmap != NULL)
        {
          int dir = enemy->mDirX*2;
          int py = player.mBitmap->Y;
          if(bitmap->X < (player.mBitmap->X + PLAYER_W/2))
          {
            dir = -dir;
          }
          
          if(py <= bitmap->Y)
            bitmap->Y += dir;
          if(py >= bitmap->Y)
            bitmap->Y -= dir;
          
          if(bitmap->Y<0)
            bitmap->Y = 0;
          if(bitmap->Y >= (FBH-PLAYER_H))
          {
            bitmap->Y = FBH-PLAYER_H-1;
          }
        }
        break;
      default:
        bitmap->X -= enemy->mDirX;
      }
      if(bitmap->X < -(PLAYER_W+16))
      {
        KillSprite(enemy);
        if(gamestats.lives > 0) /* Decrease score only if the game is still going.*/
          gamestats.score -= 10 + (enemy->mType-EEnemy)*10;
      }
    }
    enemy++;
  }
}

/* Move missiles
 * IN: missile (CSpriteState*) pointer to missile sprite array.
 *      n (int) Number of sprites in array. 
 */
void MoveMissiles(CSpriteState* missile,int n)
{
  int i;
  CBitmap* bitmap;
  
  if(missile == NULL) return;
  for(i = 0; i < n; i++)
  {
    if( (missile->mState == EAlive) && ((bitmap=missile->mBitmap) != NULL) )
    {
      bitmap->X += MISSILE_SPEED;
      if(bitmap->X > FBW)
        KillSprite(missile);
    }
    missile++;
  }
}

/* Move junk
 * IN: junk (CBitmap*) pointer to junk bitmap array.
 *      n (int) number of junk bitmap. 
 */ 
void MoveJunk(CBitmap* junk, int n)
{
  int i;
  if(junk == NULL) return;
  for(i = 0; i<n; i++)
  {
    if(junk->X> (-JUNK_W*2))
    {
      junk->X--;
    }
    junk++;
  }    
}

/* Fire a missile if the player is alive.
 * ReadJoypads() must be called in main loop for this procedure to work.
 * IN: player (CSpriteState*) pointer to player sprite.
 *     missile (CSpriteState*) pointer to array of missiles.
 *     n (int) number of missiles in array.  
 */ 
void TestPlayerFire(CSpriteState* player, CSpriteState* missile, int n)
{ 
  if(player == NULL) return;  
  if((joy1cur&(1<<FIRE_B)) && ((player->mState == EAlive)))
  { 
    GenerateMissiles(player, missile, n);
  }
}

/* Respawn the player if he's dead
 * ReadJoypads() must be called in main loop for this procedure to work.
 * IN: player (CSpriteState*) pointer to player sprite.
 */ 
void TestRespawnPlayer(CSpriteState* player)
{
  CBitmap* bitmap;
  if(player == NULL) return;  
  if(gamestats.lives <= 0) return;
  
  if(player->mState == EDead)
  {
    GeneratePlayer(player);
    player->mState = EInvicible;
    player->mTimeOfInvicibility = ticks;
    bitmap = player->mBitmap;
    if(bitmap!=NULL)
    {
      bitmap->csflags &= ~(1<<ACTIVE_COLLISION_FLAG); /* Disable collision flag when invicible */
      bitmap->addr = (&playerbitmapdata)+PLAYER_W*PLAYER_H; /* Invicible anim starts right after player sprite */
      MakeAnim(&player->mAnim, bitmap, PLAYER_W*PLAYER_H*2, INVICIBLE_STEPS, INVICIBLE_TICKS);
    }
  } 
}
 
/* Test collision of player against enemies.
 * For this test to work properly, the order of enemy bitmaps in the bitmap array
 * must be the same as the order of enemy sprites in the enemy sprite array.
 * 
 * IN: player (CSpriteState*) pointer to player sprite
 *    firstEnemy (CSpriteState*) pointer to enemy sprite array.
 *    n (int) number of enemy in the enemy sprite array. 
 */
void TestPlayerVsEnemyCollision(CSpriteState* player, CSpriteState* firstEnemy, int n)
{
  CBitmap* bitmapplayer, * bitmapenemies;
  if((player == NULL) || (firstEnemy == NULL)) return;
  
  bitmapplayer = player->mBitmap;
  bitmapenemies = firstEnemy->mBitmap;
  TestBitmapCollision(bitmapplayer, bitmapenemies, n);
  if(bitmapplayer->csflags & (1<<COLLIDE_FLAG))
  {
    ExplodeSprite(player);
    gamestats.lives--;
  }
}

/* Test collision of player against junk.
 * must be the same as the order of enemy sprites in the enemy sprite array.
 * 
 * IN: player (CSpriteState*) pointer to player sprite
 *    junk (CSpriteState*) pointer to junk bitmap array.
 *    n (int) number of junk bitmaps in the junk bitmap array. 
 *    
 * BUG: un missile est identifi comme junk!  
 */
void TestPlayerVsJunkCollision(CSpriteState* player, CBitmap* junk, int n)
{
static int a = 0;
  CBitmap* bitmapplayer;
  if((player == NULL) || (junk == NULL)) return;
  
  bitmapplayer = player->mBitmap;
  
  TestBitmapCollision(bitmapplayer, junk, n);
  if(bitmapplayer->csflags & (1<<COLLIDE_FLAG))
  {    
    ExplodeSprite(player);
    gamestats.lives--;
  }
}

/* Test collision of missiles against enemies.
 * For this test to work properly, the order of enemy bitmaps in the bitmap array
 * must be the same as the order of enemy sprites in the enemy sprite array.
 * 
 * IN: missile (CSpriteState*) pointer to missile sprite array.
 *    nm (int) number of missiles in the missile sprite array.
 *    firstEnemy (CSpriteState*) pointer to enemy sprite array.
 *    ne (int) number of enemy in the enemy sprite array. 
 */
void TestMissileVsEnemyCollision(CSpriteState* missile, int nm, CSpriteState* enemy, int ne)
{
  CBitmap* bitmapmissile, * bitmapenemies;
  int i,j;
  
  if(missile == NULL) return;
  if(enemy == NULL) return;
  
  bitmapenemies = enemy->mBitmap;  
  
  for(i = 0; i< nm; i++)
  {
    if(missile->mState == EAlive)
    {
      bitmapmissile = missile->mBitmap;
      TestBitmapCollision(bitmapmissile, bitmapenemies, ne);
      if(bitmapmissile->csflags & (1<<COLLIDE_FLAG))
      {
        KillSprite(missile);
        for(j = 0; j < ne; j++)
        {
          if(bitmapenemies[j].csflags & (1<<COLLIDE_FLAG))
          {
            ExplodeSprite(&enemy[j]);
            gamestats.score += 10 + (enemy[j].mType-EEnemy)*10;
            gamestats.nbkilled++;
          }
        }
      }      
    }
    missile++;
  }
}

/* Disable collision detection on sprite and set status to dying.
 * IN: sprite (CSpriteState*) pointer to sprite.  
 */
void ExplodeSprite(CSpriteState* sprite)
{
static int chan = 0;
  CBitmap* bitmap;
  if(sprite == NULL) return;
  sprite->mState = EDying;
  sprite->mTimeOfDeath = ticks;
  bitmap = sprite->mBitmap;
  if(bitmap != NULL)
  {
    bitmap->csflags &= ~(1<<ACTIVE_COLLISION_FLAG); /* No longer test collision on this sprite. */
    bitmap->addr = &exp4832bitmapdata;
    MakeAnim(&sprite->mAnim, bitmap, PLAYER_W*PLAYER_H*2, DYING_STEPS, DYING_TICKS);
    if(sprite->mType == EPlayer)
    {
      PlaySampleDoubleChan(6, SHIPEXPLO_SPL);
    }
    else
    {
      PlaySample(6+chan, EXPLOSION_SPL);
    }
    chan = (chan +1) & 0x1;
  }
}

/* Update status of a group of sprite.
 * Handle "dying" and "Invicible" states.
 * IN: sprite (CSpriteState*) pointer to array of sprites.
 *      n (int) number of sprites in array. 
 */ 
void UpdateSprites(CSpriteState* sprite, int n)
{
  int i;
  if(sprite == NULL) return;
  
  for(i = 0; i < n; i++)
  {
    switch(sprite->mState)
    {
    case EDying:
    {
      int delta = ticks - sprite->mTimeOfDeath;
      if(delta >= DYING_TIME)
      {
        KillSprite(sprite);
      }
      else
      {
        UpdateAnim(&sprite->mAnim, 1);
      }
    }
    break;
    
    case EInvicible:
    {
      CBitmap* bitmap;
      int delta = ticks - sprite->mTimeOfInvicibility;
      if(delta > PLAYER_INVICIBLE_TIME)
      {
        sprite->mState = EAlive;
        bitmap = sprite->mBitmap;
        if(bitmap!=NULL)
        {
          bitmap->addr = &playerbitmapdata;
          bitmap->csflags |= (1<<ACTIVE_COLLISION_FLAG); /* Enable collision flag */
        }
      }
      else
      {
        UpdateAnim(&sprite->mAnim, 1);
      }
    }
    break;
    
    default:
      ;
    }
    sprite++;
  }
}

/* Disable collision detection on sprite and hide it
 * Set sprite status to EDead
 * IN: sprite (CSpriteState*) pointer to sprite.  
 */
void KillSprite(CSpriteState* sprite)
{
  CBitmap* bitmap;
  if(sprite == NULL) return;
  sprite->mState = EDead;
  bitmap = sprite->mBitmap;
  if(bitmap != NULL)
  {
    bitmap->csflags &= ~(1<<ACTIVE_COLLISION_FLAG); /* No longer test collision on this sprite. */
    bitmap->csflags &= ~(1<<COLLIDE_FLAG);      /* reset collide flag. */
    bitmap->Y = 300; /* Hide the sprite */    
  }
}

/* Enable collision detection on player and set it at default position
 * IN: player (CSpriteState*) pointer to the player sprite. 
 */
void GeneratePlayer(CSpriteState* player)
{
  CBitmap* bitmap;
  if (player == NULL) return;
  bitmap = player->mBitmap; 
  if (bitmap==NULL) return;
  player->mState = EAlive;
  bitmap->X = 0;
  bitmap->Y = (FBH-PLAYER_H)/2;
  bitmap->csflags |= (1<<ACTIVE_COLLISION_FLAG); /* Enable test collision on this sprite. */
  bitmap->csflags &= ~(1<<COLLIDE_FLAG);      /* reset collide flag. */
  bitmap->addr = &playerbitmapdata;
}

/* Generate enemies: 
 * Enable collision detection and makes them appear at the right side of the screen
 * Only dead previously dead ennemies are generated.
 * Static variables lastgen & gendelta are used to time the generation of enemies.
 *
 * IN: enemy (CSpriteState*)  Pointer to the array of enemies.
 *     n (int)               Number of enemies in array.
 */
void GenerateEnemies(CSpriteState* enemy, int n)
{
static int lastgen = 0;
static int gendelta = 5;  
static int geny = 0;
  CBitmap* bitmap;
  int i;
  if(enemy == NULL) return;
    
  for(i = 0; i < n; i++)
  {
    if((enemy->mState == EDead)&&((ticks - lastgen)>gendelta)&&(player.mState == EAlive))
    {
      int tmp = rand();
      int acceleration = 0;
      int typeoffset = (tmp>>8)&0x3;
      int dlevel = gamestats.level-4;
      
      if((gamestats.level < 5)&&(typeoffset>(gamestats.level-1)))
      {
        typeoffset = gamestats.level-1;
      }
      if(dlevel>0)
      {
        acceleration = (dlevel)>>1;
        if(acceleration > ACCELERATION_MAX) acceleration = ACCELERATION_MAX;
        gendelta = ENEMY_DELTA - dlevel*2;
        if(gendelta < ENEMY_DELTA_MIN) gendelta = ENEMY_DELTA_MIN;
      }
      else
      {
        gendelta = ENEMY_DELTA;
      }
      
      enemy->mState = EAlive;
      bitmap = enemy->mBitmap;
      lastgen = ticks;      
      enemy->mType = EEnemy + typeoffset;
      if(bitmap != NULL)
      {        
        geny = (tmp & 0xFF);        
        if(geny>(FBH-PLAYER_H-1)) geny-= (FBH-PLAYER_H-1);        
        bitmap->X = FBW;
        bitmap->Y = geny;
        bitmap->csflags |= (1<<ACTIVE_COLLISION_FLAG); /* Enable test collision on this sprite. */
        bitmap->csflags &= ~(1<<COLLIDE_FLAG);      /* reset collide flag. */
        bitmap->addr = &enemybitmapdata+PLAYER_W*PLAYER_H*typeoffset;          
      }
      
      enemy->mDirX = ENEMY_SPEED+acceleration;
      if((tmp>>10)&0x1)
      {
        enemy->mDirY = ENEMY_SPEED+acceleration;
      }
      else
      {
        enemy->mDirY = -ENEMY_SPEED-acceleration;
      }
    }
    enemy++;
  }
}

/* Generate missiles:
 *    Enable detection collision on missiles and make them appear near player.
 *    Only previously "dead" missiles can be generated.
 * IN:  player (CSpriteState*) pointer to player sprite.
 *      missile (CSpriteState*) pointer to array of missiles.
 *      n (int) number of missiles in array. 
 */
void GenerateMissiles(CSpriteState* player, CSpriteState* missile, int n)
{
static int lastgen = 0;
static int gendelta = MISSILE_FREQUENCY; 
static int chan = 0;

  CBitmap* playerbitmap;
  CBitmap* missilebitmap;  
  int i;
  
  if(player == NULL) return;
  if(missile == NULL) return;
  
  playerbitmap = player->mBitmap;
  if(playerbitmap == NULL) return;
  
  for(i = 0; i < n; i++)
  {
    if((missile->mState == EDead)&&((ticks - lastgen)>gendelta))
    {
      missile->mState = EAlive;
      missilebitmap = missile->mBitmap;
      lastgen = ticks;
      gendelta = MISSILE_FREQUENCY;
      if(missilebitmap != NULL)
      {
        missilebitmap->X = playerbitmap->X + PLAYER_W - MISSILE_W;
        missilebitmap->Y = playerbitmap->Y + (PLAYER_H - MISSILE_H)/2;
        missilebitmap->csflags |= (1<<ACTIVE_COLLISION_FLAG); /* Enable test collision on this sprite. */ 
        missilebitmap->csflags &= ~(1<<COLLIDE_FLAG);      /* reset collide flag. */  
        PlaySample(4+chan, MISSILE_SPL);
        chan = (chan+1) & 0x1;     
      }    
    }
    missile++;
  }  
}

/* Generate junk
 * IN: junk (CBitmap*) pointer to junk bitmap array.
 *      n (int) number of junk bitmap. 
 */ 
void GenerateJunk(CBitmap* junk, int n)
{
static int lastgen = -JUNK_DELTA;
static int geny;
  int i, tmp, njunk,axis;    
  if(junk == NULL) return;
  if((ticks - lastgen) > JUNK_DELTA)
  {
    lastgen = ticks;
    njunk = 1;    
    tmp = rand();
    if((tmp & 0x01) == 0)
    {
      njunk++;
      tmp >>= 2;
      axis = tmp & 0x1;
      tmp >>= 1;
    }
    
    geny = (tmp & 0xFF);        
    if(geny>(FBH-JUNK_H-1)) geny-= (FBH-JUNK_H-1);
            
    for(i = 0; i<n; i++)
    {
      if(junk->X < -JUNK_W)
      {
        njunk--;
                
        if(axis)
        {
          junk->X = FBW+8+njunk*junk->W;
          junk->Y = geny;
        }
        else
        {
          junk->X = FBW+8;
          if(geny < (FBH-JUNK_H*2-1))
          {
            junk->Y = geny+njunk*junk->H;
          }
          else
          {
            junk->Y = geny-njunk*junk->H;
          }
        }      
        if(njunk == 0)
          break;
      }
      junk++;
    } 
  } 
}